"""Backward-compatible forward-model facade.

Forward model implementation now lives in ``lls.kinetic.models`` so both
forward and inverse model code are colocated.
"""

from .kinetic.models import make_compartment_model

__all__ = ["make_compartment_model"]
